#include <iostream>
#include <typeinfo>
#include <vector>

#include "classTraits.h"


// Tests for the classTraits.h file

using namespace std;


// some data types to work on

struct simpleStruct { simpleStruct(double); operator double*() {return 0;} };
struct polymorphicStruct { virtual int member(); };

enum e1 { fred, sue, bill };

void func1(double);

union u { int u1; char u2; };


// a generic test driver

template<template<class U> class Trait, class T, bool Expected>
void runTest(char const *Name)
{

  cout << Name << " == " << Expected;

  
  if( Trait<T>::value != Expected )
  {
    cout << " test failed"; 
  }
  else
  {
    cout << " test passed";
  }

  cout << endl;

}



// the test program



int main()
{

  // use the driver to invoke the macros ad nauseum and check for expected results

  runTest< isClassType, int, false>("isClassType<int>");
  runTest< isClassType, long, false>("isClassType<long>");
  runTest< isClassType, short, false>("isClassType<short>");
  runTest< isClassType, char, false>("isClassType<char>");
  runTest< isClassType, long long, false>("isClassType<long long>");
  runTest< isClassType, unsigned int, false>("isClassType<unsigned int>");
  runTest< isClassType, unsigned long, false>("isClassType<unsigned long>");
  runTest< isClassType, unsigned short, false>("isClassType<unsigned short>");
  runTest< isClassType, unsigned char, false>("isClassType<unsigned char>");
  runTest< isClassType, unsigned long long, false>("isClassType<unsigned long long>");
  runTest< isClassType, simpleStruct, true>("isClassType<simpleStruct>");
  runTest< isClassType, polymorphicStruct, true>("isClassType<polymorphicStruct>");
  runTest< isClassType, const polymorphicStruct, true>("isClassType<const polymorphicStruct>");
  runTest< isClassType, const polymorphicStruct &, false>("isClassType<const polymorphicStruct&>");
  runTest< isClassType, e1, false>("isClassType<e1>");
  runTest< isClassType, void (*)(), false>("isClassType<void (*)()>");
  runTest< isClassType, int const, false>("isClassType<int const>");
  runTest< isClassType, int *, false>("isClassType<int *>");
  runTest< isClassType, int [4], false>("isClassType<int [4]>");
  runTest< isClassType, void, false>("isClassType<void>");

  runTest< isPolymorphicType, int, false>("isPolymorphicType<int>");
  runTest< isPolymorphicType, long, false>("isPolymorphicType<long>");
  runTest< isPolymorphicType, short, false>("isPolymorphicType<short>");
  runTest< isPolymorphicType, char, false>("isPolymorphicType<char>");
  runTest< isPolymorphicType, long long, false>("isPolymorphicType<long long>");
  runTest< isPolymorphicType, unsigned int, false>("isPolymorphicType<unsigned int>");
  runTest< isPolymorphicType, unsigned long, false>("isPolymorphicType<unsigned long>");
  runTest< isPolymorphicType, unsigned short, false>("isPolymorphicType<unsigned short>");
  runTest< isPolymorphicType, unsigned char, false>("isPolymorphicType<unsigned char>");
  runTest< isPolymorphicType, unsigned long long, false>("isPolymorphicType<unsigned long long>");
  runTest< isPolymorphicType, simpleStruct, false>("isPolymorphicType<simpleStruct>");
  runTest< isPolymorphicType, polymorphicStruct, true>("isPolymorphicType<polymorphicStruct>");
  runTest< isPolymorphicType, const polymorphicStruct, true>("isPolymorphicType<const polymorphicStruct>");
  runTest< isPolymorphicType, const polymorphicStruct &, false>("isPolymorphicType<const polymorphicStruct&>");
  runTest< isPolymorphicType, e1, false>("isPolymorphicType<e1>");
  runTest< isPolymorphicType, void (*)(), false>("isPolymorphicType<void (*)()>");
  runTest< isPolymorphicType, int const, false>("isPolymorphicType<int const>");
  runTest< isPolymorphicType, int *, false>("isPolymorphicType<int *>");
  runTest< isPolymorphicType, int [4], false>("isPolymorphicType<int [4]>");
  runTest< isPolymorphicType, void, false>("isPolymorphicType<void>");

  runTest< isPodType, int, true>("isPodType<int>");
  runTest< isPodType, long, true>("isPodType<long>");
  runTest< isPodType, short, true>("isPodType<short>");
  runTest< isPodType, char, true>("isPodType<char>");
  runTest< isPodType, long long, true>("isPodType<long long>");
  runTest< isPodType, unsigned int, true>("isPodType<unsigned int>");
  runTest< isPodType, unsigned long, true>("isPodType<unsigned long>");
  runTest< isPodType, unsigned short, true>("isPodType<unsigned short>");
  runTest< isPodType, unsigned char, true>("isPodType<unsigned char>");
  runTest< isPodType, unsigned long long, true>("isPodType<unsigned long long>");
  runTest< isPodType, simpleStruct, false>("isPodType<simpleStruct>");
  runTest< isPodType, simpleStruct *, true>("isPodType<simpleStruct *>");
  runTest< isPodType, simpleStruct const *, true>("isPodType<simpleStruct const *>");
  runTest< isPodType, polymorphicStruct, false>("isPodType<polymorphicStruct>");
  runTest< isPodType, e1, true>("isPodType<e1>");
  runTest< isPodType, void (*)(), true>("isPodType<void (*)()>");
  runTest< isPodType, int const, true>("isPodType<int const>");
  runTest< isPodType, int *, true>("isPodType<int *>");
  runTest< isPodType, int [4], true>("isPodType<int [4]>");
  runTest< isPodType, void, false>("isPodType<void>");


  runTest< isPointerType, int, false>("isPointerType<int>");
  runTest< isPointerType, long, false>("isPointerType<long>");
  runTest< isPointerType, short, false>("isPointerType<short>");
  runTest< isPointerType, char, false>("isPointerType<char>");
  runTest< isPointerType, long long, false>("isPointerType<long long>");
  runTest< isPointerType, unsigned int, false>("isPointerType<unsigned int>");
  runTest< isPointerType, unsigned long, false>("isPointerType<unsigned long>");
  runTest< isPointerType, unsigned short, false>("isPointerType<unsigned short>");
  runTest< isPointerType, unsigned char, false>("isPointerType<unsigned char>");
  runTest< isPointerType, unsigned long long, false>("isPointerType<unsigned long long>");
  runTest< isPointerType, simpleStruct, false>("isPointerType<simpleStruct>");
  runTest< isPointerType, simpleStruct *, true>("isPointerType<simpleStruct *>");
  runTest< isPointerType, simpleStruct const *, true>("isPointerType<simpleStruct const *>");
  runTest< isPointerType, polymorphicStruct, false>("isPointerType<polymorphicStruct>");
  runTest< isPointerType, e1, false>("isPointerType<e1>");
  runTest< isPointerType, void (*)(), true>("isPointerType<void (*)()>");
  runTest< isPointerType, int const, false>("isPointerType<int const>");
  runTest< isPointerType, int *, true>("isPointerType<int *>");
  runTest< isPointerType, int [4], true>("isPointerType<int [4]>");
  runTest< isPointerType, void, false>("isPodType<void>");

  cout << "int, int                     : isSameType      = " << isSameType<int,int>::value <<endl;
  cout << "int, long long               : isSameType      = " << isSameType<int,long long>::value <<endl;
  cout << "simpleStruct, simpleSt       : isSameType      = " << isSameType<simpleStruct, simpleStruct>::value <<endl;
  cout << "simpleStruct, polymorp       : isSameType      = " << isSameType<simpleStruct, polymorphicStruct>::value <<endl;
  cout << "int, int const               : isSameType      = " << isSameType<int,int const>::value <<endl;
  cout << "const simpleStruct, simpleSt : isSameType      = " << isSameType<const simpleStruct, simpleStruct>::value <<endl;
  cout << "int *, int [4]               : isSameType      = " << isSameType<int*, int [4]>::value << endl;
                                                          
  cout << "int, long                    : isConvertible   = " << isConvertibleType<int, long>::value << endl;
  cout << "long, int                    : isConvertible   = " << isConvertibleType<long, int>::value << endl;
  cout << "long const, int              : isConvertible   = " << isConvertibleType<long const, int>::value << endl;
  cout << "int*, int const*             : isConvertible   = " << isConvertibleType<int*, int const*>::value << endl;
  cout << "int const*, int *            : isConvertible   = " << isConvertibleType<int const*, int*>::value << endl;
  cout << "int&, int const&             : isConvertible   = " << isConvertibleType<int&, int const&>::value << endl;
  cout << "int const&, int &            : isConvertible   = " << isConvertibleType<int const&, int&>::value << endl;
  cout << "double, simpleStruct         : isConvertible   = " << isConvertibleType<double, simpleStruct>::value << endl;
  cout << "simpleStruct, double*        : isConvertible   = " << isConvertibleType<simpleStruct, double*>::value << endl;
  cout << "int*, simpleStruct           : isConvertible   = " << isConvertibleType<int*, simpleStruct>::value << endl;
  cout << "simpleStruct, int*           : isConvertible   = " << isConvertibleType<simpleStruct, int*>::value << endl;
  cout << "int&, int                    : same<removeRef> = " << isSameType<removeReference<int&>::type, int>::value << endl;
  cout << "int&, char                   : same<removeRef> = " << isSameType<removeReference<int&>::type, char>::value << endl;
  cout << "int&, int&                   : same<removeRef> = " << isSameType<removeReference<int&>::type, int&>::value << endl;
  cout << "int , int&                   : same<removeRef> = " << isSameType<removeReference<int>::type, int&>::value << endl;


  cout << "isConstType<int> = " << isConstType<int>::value << endl;
  cout << "isConstType<int const> = " << isConstType<int const>::value << endl;

  runTest< isFunctionType, int, false>("isFunctionType<int>");
  runTest< isFunctionType, long, false>("isFunctionType<long>");
  runTest< isFunctionType, short, false>("isFunctionType<short>");
  runTest< isFunctionType, char, false>("isFunctionType<char>");
  runTest< isFunctionType, long long, false>("isFunctionType<long long>");
  runTest< isFunctionType, unsigned int, false>("isFunctionType<unsigned int>");
  runTest< isFunctionType, unsigned long, false>("isFunctionType<unsigned long>");
  runTest< isFunctionType, unsigned short, false>("isFunctionType<unsigned short>");
  runTest< isFunctionType, unsigned char, false>("isFunctionType<unsigned char>");
  runTest< isFunctionType, unsigned long long, false>("isFunctionType<unsigned long long>");
  runTest< isFunctionType, simpleStruct, false>("isFunctionType<simpleStruct>");
  runTest< isFunctionType, simpleStruct *, false>("isFunctionType<simpleStruct *>");
  runTest< isFunctionType, simpleStruct const *, false>("isFunctionType<simpleStruct const *>");
  runTest< isFunctionType, polymorphicStruct, false>("isFunctionType<polymorphicStruct>");
  runTest< isFunctionType, e1, false>("isFunctionType<e1>");
  runTest< isFunctionType, void (*)(), true>("isFunctionType<void (*)()>");
  runTest< isFunctionType, int const, false>("isFunctionType<int const>");
  runTest< isFunctionType, int *, false>("isFunctionType<int *>");
  runTest< isFunctionType, int [4], false>("isFunctionType<int [4]>");
  runTest< isFunctionType, void, false>("isFunctionType<void>");


  runTest< isSignedType, int, true>("isSignedType<int>");
  runTest< isSignedType, long, true>("isSignedType<long>");
  runTest< isSignedType, short, true>("isSignedType<short>");
  runTest< isSignedType, char, true>("isSignedType<char>");
  runTest< isSignedType, long long, true>("isSignedType<long long>");
  runTest< isSignedType, unsigned int, false>("isSignedType<unsigned int>");
  runTest< isSignedType, unsigned long, false>("isSignedType<unsigned long>");
  runTest< isSignedType, unsigned short, false>("isSignedType<unsigned short>");
  runTest< isSignedType, unsigned char, false>("isSignedType<unsigned char>");
  runTest< isSignedType, unsigned long long, false>("isSignedType<unsigned long long>");
  runTest< isSignedType, simpleStruct, false>("isSignedType<simpleStruct>");
  runTest< isSignedType, simpleStruct *, false>("isSignedType<simpleStruct *>");
  runTest< isSignedType, simpleStruct const *, false>("isSignedType<simpleStruct const *>");
  runTest< isSignedType, polymorphicStruct, false>("isSignedType<polymorphicStruct>");
  runTest< isSignedType, e1, false>("isSignedType<e1>");
  runTest< isSignedType, void (*)(), false>("isSignedType<void (*)()>");
  runTest< isSignedType, int const, true>("isSignedType<int const>");
  runTest< isSignedType, int *, false>("isSignedType<int *>");
  runTest< isSignedType, int [4], false>("isSignedType<int [4]>");
  runTest< isSignedType, void, false>("isSignedType<void>");


  runTest< isUnsignedType, int, false>("isUnsignedType<int>");
  runTest< isUnsignedType, long, false>("isUnsignedType<long>");
  runTest< isUnsignedType, short, false>("isUnsignedType<short>");
  runTest< isUnsignedType, char, false>("isUnsignedType<char>");
  runTest< isUnsignedType, long long, false>("isUnsignedType<long long>");
  runTest< isUnsignedType, unsigned int, true>("isUnsignedType<unsigned int>");
  runTest< isUnsignedType, unsigned long, true>("isUnsignedType<unsigned long>");
  runTest< isUnsignedType, unsigned short, true>("isUnsignedType<unsigned short>");
  runTest< isUnsignedType, unsigned char, true>("isUnsignedType<unsigned char>");
  runTest< isUnsignedType, unsigned long long, true>("isUnsignedType<unsigned long long>");
  runTest< isUnsignedType, simpleStruct, false>("isUnsignedType<simpleStruct>");
  runTest< isUnsignedType, simpleStruct *, false>("isUnsignedType<simpleStruct *>");
  runTest< isUnsignedType, simpleStruct const *, false>("isUnsignedType<simpleStruct const *>");
  runTest< isUnsignedType, polymorphicStruct, false>("isUnsignedType<polymorphicStruct>");
  runTest< isUnsignedType, e1, true>("isUnsignedType<e1>");
  runTest< isUnsignedType, void (*)(), false>("isUnsignedType<void (*)()>");
  runTest< isUnsignedType, int const, false>("isUnsignedType<int const>");
  runTest< isUnsignedType, int *, false>("isUnsignedType<int *>");
  runTest< isUnsignedType, int [4], false>("isUnsignedType<int [4]>");
  runTest< isUnsignedType, void, false>("isUnsignedType<void>");


  runTest< isEnumType, int, false>("isEnumType<int>");
  runTest< isEnumType, long, false>("isEnumType<long>");
  runTest< isEnumType, short, false>("isEnumType<short>");
  runTest< isEnumType, char, false>("isEnumType<char>");
  runTest< isEnumType, long long, false>("isEnumType<long long>");
  runTest< isEnumType, unsigned int, false>("isEnumType<unsigned int>");
  runTest< isEnumType, unsigned long, false>("isEnumType<unsigned long>");
  runTest< isEnumType, unsigned short, false>("isEnumType<unsigned short>");
  runTest< isEnumType, unsigned char, false>("isEnumType<unsigned char>");
  runTest< isEnumType, unsigned long long, false>("isEnumType<unsigned long long>");
  runTest< isEnumType, simpleStruct, false>("isEnumType<simpleStruct>");
  runTest< isEnumType, simpleStruct *, false>("isEnumType<simpleStruct *>");
  runTest< isEnumType, simpleStruct const *, false>("isEnumType<simpleStruct const *>");
  runTest< isEnumType, polymorphicStruct, false>("isEnumType<polymorphicStruct>");
  runTest< isEnumType, e1, true>("isEnumType<e1>");
  runTest< isEnumType, void (*)(), false>("isEnumType<void (*)()>");
  runTest< isEnumType, int const, false>("isEnumType<int const>");
  runTest< isEnumType, int *, false>("isEnumType<int *>");
  runTest< isEnumType, int [4], false>("isEnumType<int [4]>");
  runTest< isEnumType, void, false>("isEnumType<void>");

  runTest< isArithmeticType, int, true>("isArithmeticType<int>");
  runTest< isArithmeticType, long, true>("isArithmeticType<long>");
  runTest< isArithmeticType, short, true>("isArithmeticType<short>");
  runTest< isArithmeticType, char, true>("isArithmeticType<char>");
  runTest< isArithmeticType, long long, true>("isArithmeticType<long long>");
  runTest< isArithmeticType, unsigned int, true>("isArithmeticType<unsigned int>");
  runTest< isArithmeticType, unsigned long, true>("isArithmeticType<unsigned long>");
  runTest< isArithmeticType, unsigned short, true>("isArithmeticType<unsigned short>");
  runTest< isArithmeticType, unsigned char, true>("isArithmeticType<unsigned char>");
  runTest< isArithmeticType, unsigned long long, true>("isArithmeticType<unsigned long long>");
  runTest< isArithmeticType, simpleStruct, false>("isArithmeticType<simpleStruct>");
  runTest< isArithmeticType, simpleStruct *, false>("isArithmeticType<simpleStruct *>");
  runTest< isArithmeticType, simpleStruct const *, false>("isArithmeticType<simpleStruct const *>");
  runTest< isArithmeticType, polymorphicStruct, false>("isArithmeticType<polymorphicStruct>");
  runTest< isArithmeticType, e1, false>("isArithmeticType<e1>");
  runTest< isArithmeticType, void (*)(), false>("isArithmeticType<void (*)()>");
  runTest< isArithmeticType, int const, true>("isArithmeticType<int const>");
  runTest< isArithmeticType, int *, false>("isArithmeticType<int *>");
  runTest< isArithmeticType, int [4], false>("isArithmeticType<int [4]>");
  runTest< isArithmeticType, void, false>("isArithmeticType<void>");
  runTest< isArithmeticType, float, true>("isArithmeticType<float>");
  runTest< isArithmeticType, double, true>("isArithmeticType<double>");


  runTest< isIntegralType, int, true>("isIntegralType<int>");
  runTest< isIntegralType, long, true>("isIntegralType<long>");
  runTest< isIntegralType, short, true>("isIntegralType<short>");
  runTest< isIntegralType, char, true>("isIntegralType<char>");
  runTest< isIntegralType, long long, true>("isIntegralType<long long>");
  runTest< isIntegralType, unsigned int, true>("isIntegralType<unsigned int>");
  runTest< isIntegralType, unsigned long, true>("isIntegralType<unsigned long>");
  runTest< isIntegralType, unsigned short, true>("isIntegralType<unsigned short>");
  runTest< isIntegralType, unsigned char, true>("isIntegralType<unsigned char>");
  runTest< isIntegralType, unsigned long long, true>("isIntegralType<unsigned long long>");
  runTest< isIntegralType, simpleStruct, false>("isIntegralType<simpleStruct>");
  runTest< isIntegralType, simpleStruct *, false>("isIntegralType<simpleStruct *>");
  runTest< isIntegralType, simpleStruct const *, false>("isIntegralType<simpleStruct const *>");
  runTest< isIntegralType, polymorphicStruct, false>("isIntegralType<polymorphicStruct>");
  runTest< isIntegralType, e1, false>("isIntegralType<e1>");
  runTest< isIntegralType, void (*)(), false>("isIntegralType<void (*)()>");
  runTest< isIntegralType, int const, true>("isIntegralType<int const>");
  runTest< isIntegralType, int *, false>("isIntegralType<int *>");
  runTest< isIntegralType, int [4], false>("isIntegralType<int [4]>");
  runTest< isIntegralType, void, false>("isIntegralType<void>");
  runTest< isIntegralType, float, false>("isIntegralType<float>");
  runTest< isIntegralType, double, false>("isIntegralType<double>");

  removeReference< int &>::type *p = 0; // no output, just give a compile error if it fails

  cout << "isTemplate<int>                    : " << isTemplate<int>::value << endl;
  cout << "isTemplate<vector<int>>            : " << isTemplate<vector<int> >::value << endl;
  cout << "isTemplate<const vector<const int>>: " << isTemplate<const vector< const int> >::value << endl;

}




